/*
 *   jack-ladspa-host
 *    
 *   Copyright (C) Robert Ham 2002 (node@users.sourceforge.net)
 *    
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
#define _GNU_SOURCE

#include "ac_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>

#include <gtk/gtk.h>
#include <ladspa.h>

#ifdef HAVE_GNOME
#include <libgnomeui/libgnomeui.h>
#endif

#include "jack_rack.h"
#include "lock_free_fifo.h"
#include "callbacks.h"
#include "control_message.h"
#include "globals.h"
#include "file.h"


void
add_callback (GtkMenuItem * menuitem, gpointer user_data)
{
  jack_rack_t * jack_rack;
  plugin_desc_t * desc;
  plugin_t * plugin;
  ctrlmsg_t ctrlmsg;
  
  jack_rack = (jack_rack_t *) user_data;
  
  /* get the selected plugin out of the gobject data thing */
  desc = g_object_get_data (G_OBJECT(menuitem), "jack-rack-plugin-desc");
  if (!desc)
  {
    fprintf (stderr, "%s: no plugin description!\n", __FUNCTION__);
    return;
  }
  
  plugin = jack_rack_instantiate_plugin (jack_rack, desc);

  if (!plugin)
    return;
    
  
  /* send the chain link off to the process() callback */
  ctrlmsg.type = CTRLMSG_ADD;
  ctrlmsg.pointer = plugin;
  ctrlmsg.second_pointer = desc;
  lff_write (jack_rack->gui_to_process, &ctrlmsg);

}

void
new_callback (GtkWidget * widget, gpointer user_data)
{
  jack_rack_t * jack_rack;
  ctrlmsg_t ctrlmsg;
  
  jack_rack = (jack_rack_t *) user_data;
  
  ctrlmsg.type = CTRLMSG_CLEAR;
  
  lff_write (jack_rack->gui_to_process, &ctrlmsg);
}

static const char *
get_filename (jack_rack_t * jack_rack)
{
  static GtkWidget * dialog = NULL;
  static char * file = NULL;
  int response;
  
  if (!dialog)
    dialog = gtk_file_selection_new ("Select file");
  
  
  response = gtk_dialog_run (GTK_DIALOG (dialog));
  gtk_widget_hide (dialog);
  
  if (response != GTK_RESPONSE_OK)
    return NULL;

  if (file)
    g_free (file);
  
  file = g_strdup (gtk_file_selection_get_filename (GTK_FILE_SELECTION (dialog)));
  return file;
}



void
open_callback (GtkButton * button, gpointer user_data)
{
  const char * filename;
  jack_rack_t * jack_rack;
  
  jack_rack = (jack_rack_t *) user_data;
  
  filename = get_filename (jack_rack);
  
  if (!filename)
    return;
  
  jack_rack_open_file (jack_rack, filename);
}
 
void
save_callback (GtkButton * button, gpointer user_data)
{
  jack_rack_t * jack_rack;
  
  jack_rack = (jack_rack_t *) user_data;
  
  if (!jack_rack->filename)
    {
      save_as_callback (button, user_data);
      return;
    }
  
  jack_rack_save_file (jack_rack, jack_rack->filename);
}


 
void
save_as_callback (GtkButton * button, gpointer user_data)
{
  const char * filename;
  jack_rack_t * jack_rack;
  
  jack_rack = (jack_rack_t *) user_data;
  
  filename = get_filename (jack_rack);
  
  if (!filename)
    return;
  
  jack_rack_save_file (jack_rack, filename);
}
 
#ifdef HAVE_LADCCA
void
cca_save_callback (GtkButton * button, gpointer user_data)
{
  cca_event_t * event;
  
  printf ("%s: sending event\n", __FUNCTION__);
  event = cca_event_new_with_type (CCA_Save);
  cca_send_event (global_cca_client, event);
  printf ("%s: event sent\n", __FUNCTION__);
}
#endif


void
quit_callback (GtkButton * button, gpointer user_data)
{
  jack_rack_t * jack_rack;
  ctrlmsg_t ctrlmsg;
  
  jack_rack = user_data;
  
  ctrlmsg.type = CTRLMSG_CLEAR;
  lff_write (jack_rack->gui_to_process, &ctrlmsg);

  ctrlmsg.type = CTRLMSG_QUIT;
  lff_write (jack_rack->gui_to_process, &ctrlmsg);
}

gboolean
window_destroy_callback (GtkWidget *widget,
                         GdkEvent *event,
                         gpointer user_data)
{
  quit_callback (NULL, user_data);
  return TRUE;
}

#ifdef HAVE_GNOME
void
about_callback (GtkWidget * widget, gpointer user_data)
{
  GtkWidget * dialog;
  const char * authors[] = { "Bob Ham <node@users.sourceforge.net>", NULL };
  const char * documenters[] = { NULL };
  GtkWidget * url;
  gchar * logo_file;
  GdkPixbuf * logo;
  GError * err = NULL;
  
  logo_file = g_strdup_printf ("%s/%s", PKGDATADIR, JACK_RACK_LOGO_FILE);
  logo = gdk_pixbuf_new_from_file (logo_file, &err);
  if (err)
    {
      fprintf (stderr, "%s: error loading logo image from file '%s': '%s'\n",
               __FUNCTION__, logo_file, err->message);
      g_error_free (err);
    }
  g_free (logo_file);
  
  dialog = gnome_about_new (PACKAGE_NAME,
                              PACKAGE_VERSION,
                              "Copyright (C) 2002, 2003 Robert Ham <node@users.sourceforge.net>",
                              "A stereo LADSPA effects rack for the JACK audio API",
                              authors,
                              documenters,
                              "",
                              logo);
  
  url = gnome_href_new (JACK_RACK_URL, JACK_RACK_URL);
  gtk_widget_show (url);
  GTK_WIDGET_UNSET_FLAGS (url, GTK_CAN_FOCUS);
  gtk_box_pack_start (GTK_BOX (GTK_DIALOG(dialog)->vbox), url, FALSE, FALSE, 0);
  
  gtk_dialog_run (GTK_DIALOG (dialog));
}
#endif /* HAVE_GNOME */

void
slot_change_callback (GtkMenuItem * menuitem, gpointer user_data)
{
  ctrlmsg_t ctrlmsg;
  plugin_slot_t * plugin_slot;
  plugin_t * plugin;
  plugin_desc_t * desc;
  
  plugin_slot = user_data;
  
  if (global_jack_rack->state == STATE_RACK_CHANGE)
    return;
  
  desc = (plugin_desc_t *) g_object_get_data (G_OBJECT(menuitem), "jack-rack-plugin-desc");

  plugin = jack_rack_instantiate_plugin (global_jack_rack, desc);
  
  if (!plugin)
    return;

  ctrlmsg.type = CTRLMSG_CHANGE;
  ctrlmsg.number = plugin_slot->index;
  ctrlmsg.pointer = plugin;
  ctrlmsg.second_pointer = plugin;
  
  lff_write (global_jack_rack->gui_to_process, &ctrlmsg);
}

void
slot_move_callback (GtkButton * button, gpointer user_data)
{
  ctrlmsg_t ctrlmsg;
  plugin_slot_t * plugin_slot;
  gint up;

  plugin_slot = user_data;
  up = GPOINTER_TO_INT (g_object_get_data (G_OBJECT(button), "jack-rack-up"));
  
  /* this is because we use indicies */
  if (global_jack_rack->state == STATE_RACK_CHANGE)
    return;
  
  /* check the logic of what we're doing */
  if ((up && !plugin_slot->prev) ||
      (!up && !plugin_slot->next))
    return;
  
  jack_rack_set_state (global_jack_rack, STATE_RACK_CHANGE);
  
  ctrlmsg.type = CTRLMSG_MOVE;
  ctrlmsg.number = plugin_slot->index;
  ctrlmsg.pointer = GINT_TO_POINTER (up);
  ctrlmsg.second_pointer = plugin_slot;
  
  lff_write (global_jack_rack->gui_to_process, &ctrlmsg);
  
}

void slot_remove_callback (GtkButton * button, gpointer user_data) {
  ctrlmsg_t ctrlmsg;
  plugin_slot_t * plugin_slot;
  
  plugin_slot = (plugin_slot_t *) user_data;
  
  ctrlmsg.type = CTRLMSG_REMOVE;
  ctrlmsg.number = plugin_slot->index;
  ctrlmsg.second_pointer = plugin_slot;
  
  lff_write (global_jack_rack->gui_to_process, &ctrlmsg);
}


gboolean slot_ablise_callback (GtkWidget * button, GdkEventButton *event, gpointer user_data) {
  if (event->type == GDK_BUTTON_PRESS) {
    plugin_slot_t * plugin_slot; 
    ctrlmsg_t ctrlmsg;
    gint ablise;
  
    plugin_slot = (plugin_slot_t *) user_data;
    ablise = (plugin_slot->enabled ? FALSE : TRUE);
  
    ctrlmsg.type = CTRLMSG_ABLE;
    ctrlmsg.number = plugin_slot->index;
    ctrlmsg.pointer = GINT_TO_POINTER(ablise);
    ctrlmsg.second_pointer = plugin_slot;
  
    lff_write (global_jack_rack->gui_to_process, &ctrlmsg);
  
    return TRUE;
  }
  
  return FALSE;
}

void slot_lock_all_callback (GtkToggleButton * button, gpointer user_data) {
  gboolean on;
  plugin_slot_t * plugin_slot;
  unsigned long i;
  
  on = gtk_toggle_button_get_active (button);
  
  plugin_slot = (plugin_slot_t *) user_data;
  
/*  plugin_slot->plugin_settings[plugin_slot->plugin->collection_index].lock_all = on; */
  settings_set_lock_all (plugin_slot->settings, on);

  for (i = 0; i < plugin_slot->plugin->desc->control_port_count; i++) {
  
    gtk_widget_set_sensitive (plugin_slot->port_controls[i].lock, on ? FALSE : TRUE);
      
    if (on) {
      plugin_slot->port_controls[i].locked = TRUE;
    } else {
      plugin_slot->port_controls[i].locked =
        gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON(plugin_slot->port_controls[i].lock));
    }
    
  }
  
}

void control_lock_callback (GtkToggleButton * button, gpointer user_data) {
  port_controls_t * port_controls;
  
  port_controls = (port_controls_t *) user_data;
  
  port_controls->locked = gtk_toggle_button_get_active (button);
  settings_set_lock (port_controls->plugin_slot->settings, port_controls->control_index, port_controls->locked);

/*  port_controls_holder->plugin_slot->
    plugin_settings[port_controls_holder->plugin_slot->plugin->collection_index].
    locks[port_controls_holder->port_index] = port_controls_holder->locked; */
}

void set_widget_text_colour (GtkWidget * widget,
                         const char * colour_str) {
  GdkColor colour;
  GdkColormap * colourmap;
  
  colourmap = gtk_widget_get_colormap (widget);
  
  gdk_color_parse (colour_str, &colour);
  gdk_colormap_alloc_color (colourmap, &colour, FALSE, TRUE);
  gtk_widget_modify_text (widget, GTK_STATE_NORMAL, &colour);
}

void control_float_callback (GtkRange * range, gpointer user_data) {
  port_controls_t * port_controls;
  gchar * str;
  LADSPA_Data value;
  int copy;
  
  port_controls = (port_controls_t *) user_data;
  
  value = gtk_range_get_value (range);
  
  str = g_strdup_printf ("%f", value);
  
  /* get which copy we're using from the g object data stuff */
  copy = GPOINTER_TO_INT (g_object_get_data (G_OBJECT(range), "jack-rack-plugin-copy"));
    
  /* write to the fifo */
  lff_write (port_controls->controls[copy].control_fifo, &value);
    
  /* print the value in the text box */
  gtk_entry_set_text (GTK_ENTRY(port_controls->controls[copy].text), str);
    
  /* store the value */
  settings_set_control_value (port_controls->plugin_slot->settings, port_controls->control_index, copy, value);
              
    
  /* possibly set our pair */
  if (port_controls->plugin_slot->plugin->desc->mono && port_controls->locked) {
    unsigned long i;
    for (i = 0; i < plugin_copies (port_controls->plugin_slot->plugin->desc); i++)
      if (i != copy)
        gtk_range_set_value (GTK_RANGE(port_controls->controls[i].control),
                             gtk_range_get_value (range));
  }
    
    
  set_widget_text_colour (port_controls->controls[copy].text, "Black");
  
  g_free (str);
  
}

void control_float_text_callback (GtkEntry * entry, gpointer user_data) {
  port_controls_t * port_controls;
  LADSPA_Data value;
  const char * str;
  char * endptr;
  GtkAdjustment * adjustment;
  gint copy;
  
  port_controls = (port_controls_t *) user_data;
  copy = GPOINTER_TO_INT (g_object_get_data (G_OBJECT(entry), "jack-rack-plugin-copy"));

  str = gtk_entry_get_text (entry);
  value = strtof (str, &endptr);
  
  /* check for errors */
  adjustment = gtk_range_get_adjustment (GTK_RANGE (port_controls->controls[copy].control));

  if (value == HUGE_VALF ||
      value == -HUGE_VALF ||
      endptr == str ||
      value < adjustment->lower ||
      value > adjustment->upper) {
    /* set the entry's text colour */
    set_widget_text_colour (GTK_WIDGET(entry), "DarkRed");
    return;
  }
  
  
  /* set the value in range */
  gtk_range_set_value (GTK_RANGE(port_controls->controls[copy].control), value);
    
  /* possibly set our peers */
  if (port_controls->plugin_slot->plugin->desc->mono && port_controls->locked) {
    unsigned long i;
    for (i = 0; i < plugin_copies (port_controls->plugin_slot->plugin->desc); i++)
      if (i != copy)
        gtk_range_set_value (GTK_RANGE(port_controls->controls[i].control), value);
  }
  
  set_widget_text_colour (GTK_WIDGET(entry), "Black");
}


void control_bool_callback (GtkToggleButton * button, gpointer user_data) {
  port_controls_t * port_controls;
  gboolean on;
  LADSPA_Data data;
  int copy;
  
  port_controls = (port_controls_t *) user_data;
  
  on = gtk_toggle_button_get_active (button);
  data = on ? 1.0 : -1.0;

  /* get which channel we're using from the g object data stuff */
  copy = GPOINTER_TO_INT (g_object_get_data (G_OBJECT(button), "jack-rack-plugin-copy"));
  
  /* write to the fifo */
  lff_write (port_controls->controls[copy].control_fifo, &data);
  
  /* store value */
  settings_set_control_value (port_controls->plugin_slot->settings, port_controls->control_index, copy, data);
  
  /* possibly set other controls */
  if (port_controls->plugin_slot->plugin->desc->mono && port_controls->locked) {
    unsigned long i;
    for (i = 0; i < plugin_copies (port_controls->plugin_slot->plugin->desc); i++)
      if (i != copy)
        gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(port_controls->controls[i].control), on);
    }
}


void control_int_callback (GtkSpinButton * spinbutton, gpointer user_data) {
  port_controls_t * port_controls;
  int value;
  LADSPA_Data data;
  int copy;
  
  port_controls = (port_controls_t *) user_data;
  
  value = gtk_spin_button_get_value_as_int (spinbutton);
  data = value;
  
  /* get which channel we're using from the g object data stuff */
  copy = GPOINTER_TO_INT (g_object_get_data (G_OBJECT(spinbutton), "jack-rack-plugin-copy"));
  
  /* write to the fifo */
  lff_write (port_controls->controls[copy].control_fifo, &data);
  
  /* store value */
  settings_set_control_value (port_controls->plugin_slot->settings, port_controls->control_index, copy, data);
      
  /* possibly set other controls */
  if (port_controls->plugin_slot->plugin->desc->mono && port_controls->locked) {
    unsigned long i;
    for (i = 0; i < plugin_copies (port_controls->plugin_slot->plugin->desc); i++)
      if (i != copy)
        {
          gtk_spin_button_set_value (GTK_SPIN_BUTTON (port_controls->controls[copy].control), value);
        }
    }
}

/** callback for plugin menu buttons (Add and the plugin slots' change buttons) */
gint plugin_button_callback (GtkWidget *widget, GdkEvent *event) {

  if (event->type == GDK_BUTTON_PRESS) {
    GdkEventButton * bevent = (GdkEventButton *) event;
    gtk_menu_popup (GTK_MENU (widget), NULL, NULL, NULL, NULL,
                    bevent->button, bevent->time);
    /* Tell calling code that we have handled this event; the buck
     * stops here. */
    return TRUE;
  }
  
  /* Tell calling code that we have not handled this event; pass it on. */
  return FALSE;
}

#ifdef HAVE_LADCCA
void
deal_with_cca_event (jack_rack_t * jack_rack, cca_event_t * event)
{
  switch (cca_event_get_type (event))
    {
    case CCA_Save_File:
      jack_rack_save_file (jack_rack, cca_get_fqn (cca_event_get_string (event), "jack_rack.rack"));
      cca_send_event (global_cca_client, event);
      break;
    case CCA_Restore_File:
      jack_rack_open_file (jack_rack, cca_get_fqn (cca_event_get_string (event), "jack_rack.rack"));
      cca_send_event (global_cca_client, event);
      break;
    case CCA_Quit:
      quit_callback (NULL, jack_rack);
      cca_event_destroy (event);
      break;
    default:
      fprintf (stderr, "Recieved LADCCA event of unknown type %d\n", cca_event_get_type (event));
      cca_event_destroy (event);
      break;
    }
}
#endif /* HAVE_LADCCA */

static void
jack_rack_check_kicked (jack_rack_t * jack_rack)
{
  static gboolean shown_shutdown_msg = FALSE;
  if (jack_rack->shutdown && !shown_shutdown_msg)
  {
    jack_rack_display_error (jack_rack, "JACK client thread shut down by server\n\nJACK, the bastard, kicked us out.  You'll have to restart I'm afraid.  Sorry.");
    
    gtk_widget_set_sensitive (jack_rack->plugin_box, FALSE);
    gtk_widget_set_sensitive (jack_rack->add, FALSE);

    shown_shutdown_msg = TRUE;
  }
}

/* do the process->gui message processing */
gboolean idle_callback (gpointer data) {
  ctrlmsg_t ctrlmsg;
  jack_rack_t * jack_rack;
  plugin_t * plugin;
  plugin_slot_t * plugin_slot;
  int enabled;

#ifdef HAVE_LADCCA
  static int ladcca_enabled_at_start = -10;
  
  if (ladcca_enabled_at_start == -10)
    ladcca_enabled_at_start = cca_enabled (global_cca_client);
#endif


  jack_rack = (jack_rack_t *) data;
  
  jack_rack_check_kicked (jack_rack);
  
  while (lff_read (jack_rack->process_to_gui, &ctrlmsg) == 0)
    {
    switch (ctrlmsg.type)
      {
      case CTRLMSG_ADD:
        plugin = ctrlmsg.pointer;
        jack_rack_add_plugin_slot (jack_rack, plugin);
        break;

      case CTRLMSG_REMOVE:
        plugin = ctrlmsg.pointer;
        plugin_slot = ctrlmsg.second_pointer;
        jack_rack_remove_plugin_slot (jack_rack, plugin_slot);
        plugin_destroy (plugin);
        break;

      case CTRLMSG_MOVE:
        plugin_slot = ctrlmsg.second_pointer;
        jack_rack_move_plugin_slot (jack_rack, plugin_slot, GPOINTER_TO_INT (ctrlmsg.pointer));
        jack_rack_set_state (jack_rack, STATE_NORMAL);
        break;

      case CTRLMSG_CHANGE:
        plugin_slot = jack_rack_get_plugin_slot (jack_rack, ctrlmsg.number);
        plugin = ctrlmsg.pointer;
        jack_rack_change_plugin_slot (jack_rack, plugin_slot, ctrlmsg.second_pointer);
        plugin_destroy (plugin);
        break;

      case CTRLMSG_ABLE:
        plugin_slot = ctrlmsg.second_pointer;
        enabled = GPOINTER_TO_INT(ctrlmsg.pointer);
        gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(plugin_slot->enable),
                                      enabled);
        plugin_slot->enabled = enabled;
        break;
      
      case CTRLMSG_CLEAR:
        plugin = ctrlmsg.pointer;
        jack_rack_clear_plugins (jack_rack, plugin);
        break;

      case CTRLMSG_QUIT: /* we'll never get a QUIT */
        gtk_main_quit ();
        break;
      }
    }

#ifdef HAVE_LADCCA
  if (ladcca_enabled_at_start && !cca_enabled (global_cca_client))
    {
      /* server disconnected */
      printf ("ladcca server disconnected\n");
      gtk_widget_set_sensitive (jack_rack->cca_save, FALSE);
      gtk_widget_set_sensitive (jack_rack->cca_save_menu_item, FALSE);
      ladcca_enabled_at_start = 0;
    }
  
  if (ladcca_enabled_at_start && cca_enabled (global_cca_client))
    {
      cca_event_t * event;
      
      while ( (event = cca_get_event (global_cca_client)) )
        {
          deal_with_cca_event (jack_rack, event);
        }
    }
#endif
  
  usleep (1000);
  
  return TRUE;
}

void jack_shutdown_callback (void * data) {
  jack_rack_t * jack_rack;
  
  jack_rack = data;
  
  jack_rack->shutdown = TRUE;
}

