/*
 *   JACK Rack
 *    
 *   Copyright (C) Robert Ham 2002, 2003 (node@users.sourceforge.net)
 *    
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "ac_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <string.h>
#include <ctype.h>

#include <gtk/gtk.h>
#include <ladspa.h>

#ifdef HAVE_LADCCA
#include <ladcca/ladcca.h>
#endif

#include "jack_rack.h"
#include "lock_free_fifo.h"
#include "callbacks.h"
#include "control_message.h"
#include "callbacks.h"
#include "globals.h"

static void
jack_rack_init_gui_menu (jack_rack_t * rack, GtkWidget * main_box)
{
  GtkWidget *menubar_handle;
  GtkWidget *menubar;
  GtkWidget *file_menuitem;
  GtkWidget *add_menuitem;
  GtkWidget *file_menu;
  GtkWidget *add_menu;
  GtkWidget *new;
  GtkWidget *open;
  GtkWidget *save;
  GtkWidget *save_as;
  GtkWidget *separator;
  GtkWidget *quit;
#ifdef HAVE_GNOME
  GtkWidget *help_menuitem;
  GtkWidget *help_menu;
  GtkWidget *about;
#endif

  /* the menu bar */
  menubar_handle = gtk_handle_box_new ();
  gtk_widget_show (menubar_handle);
  gtk_box_pack_start (GTK_BOX (main_box), menubar_handle, FALSE, TRUE, 0);
  
  menubar = gtk_menu_bar_new ();
  gtk_widget_show (menubar);
  gtk_container_add (GTK_CONTAINER (menubar_handle), menubar);
  
  file_menuitem = gtk_menu_item_new_with_label ("File");
  gtk_widget_show (file_menuitem);
  gtk_menu_shell_append (GTK_MENU_SHELL (menubar), file_menuitem);
  
  add_menuitem = gtk_menu_item_new_with_label ("Add");
  gtk_widget_show (add_menuitem);
  gtk_menu_shell_append (GTK_MENU_SHELL (menubar), add_menuitem);

#ifdef HAVE_GNOME  
  help_menuitem = gtk_menu_item_new_with_label ("Help");
  gtk_widget_show (help_menuitem);
  gtk_menu_shell_append (GTK_MENU_SHELL (menubar), help_menuitem);
#endif

  /* file menu */
  file_menu = gtk_menu_new ();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM (file_menuitem), file_menu);
  
  new = gtk_image_menu_item_new_from_stock (GTK_STOCK_NEW, NULL);
  gtk_widget_show (new);
  gtk_menu_shell_append (GTK_MENU_SHELL (file_menu), new);
  g_signal_connect (G_OBJECT (new), "activate",
                    G_CALLBACK (new_callback), rack);

  open = gtk_image_menu_item_new_from_stock (GTK_STOCK_OPEN, NULL);
  gtk_widget_show (open);
  gtk_menu_shell_append (GTK_MENU_SHELL (file_menu), open);
  g_signal_connect (G_OBJECT (open), "activate",
                    G_CALLBACK (open_callback), rack);

  save = gtk_image_menu_item_new_from_stock (GTK_STOCK_SAVE, NULL);
  gtk_widget_show (save);
  gtk_menu_shell_append (GTK_MENU_SHELL (file_menu), save);
  g_signal_connect (G_OBJECT (save), "activate",
                    G_CALLBACK (save_callback), rack);

  save_as = gtk_image_menu_item_new_from_stock (GTK_STOCK_SAVE_AS, NULL);
  gtk_widget_show (save_as);
  gtk_menu_shell_append (GTK_MENU_SHELL (file_menu), save_as);
  g_signal_connect (G_OBJECT (save_as), "activate",
                    G_CALLBACK (save_as_callback), rack);

#ifdef HAVE_LADCCA
  if (cca_enabled (global_cca_client))
    {
      separator = gtk_separator_menu_item_new ();
      gtk_widget_show (separator);
      gtk_menu_shell_append (GTK_MENU_SHELL (file_menu), separator);

      rack->cca_save_menu_item = gtk_menu_item_new_with_label ("Save project");
      gtk_widget_show (rack->cca_save_menu_item);
      gtk_menu_shell_append (GTK_MENU_SHELL (file_menu), rack->cca_save_menu_item);
      g_signal_connect (G_OBJECT (rack->cca_save_menu_item), "activate",
                        G_CALLBACK (cca_save_callback), rack);
    }
#endif /* HAVE_LADCCA */

  separator = gtk_separator_menu_item_new ();
  gtk_widget_show (separator);
  gtk_menu_shell_append (GTK_MENU_SHELL (file_menu), separator);
  
  quit = gtk_image_menu_item_new_from_stock (GTK_STOCK_QUIT, NULL);
  gtk_widget_show (quit);
  gtk_menu_shell_append (GTK_MENU_SHELL (file_menu), quit);
  g_signal_connect (G_OBJECT (quit), "activate",
                    G_CALLBACK (quit_callback), rack);


  /* add menu */
  add_menu = plugin_mgr_get_menu (rack->plugin_mgr, G_CALLBACK (add_callback), rack);
  gtk_menu_item_set_submenu (GTK_MENU_ITEM(add_menuitem), add_menu);


#ifdef HAVE_GNOME
  /* help menu */
  help_menu = gtk_menu_new ();
  gtk_menu_item_set_submenu (GTK_MENU_ITEM(help_menuitem), help_menu);
  
  about = gtk_menu_item_new_with_label ("About");
  gtk_widget_show (about);
  gtk_menu_shell_append (GTK_MENU_SHELL (help_menu), about);
  g_signal_connect (G_OBJECT (about), "activate",
                    G_CALLBACK (about_callback), rack);
#endif
}

static void
jack_rack_init_gui (jack_rack_t * rack)
{
  GtkWidget *menu;
  GtkWidget *main_box;
  GtkWidget *toolbar_handle;
  GtkWidget *toolbar;
  GtkWidget *plugin_scroll;
  GtkWidget *plugin_viewport;
  gchar * icon_file;
  GError * icon_error = NULL;

  /* set the icons */
  icon_file = g_strdup_printf ("%s/pixmaps/%s", JR_DESKTOP_PREFIX, JACK_RACK_ICON_FILE);
  gtk_window_set_default_icon_from_file (icon_file, &icon_error);
  if (icon_error)
    {
      fprintf (stderr, "%s: error loading icon file '%s': %s\n", __FUNCTION__, icon_file, icon_error->message);
      g_error_free (icon_error);
    }
  g_free (icon_file);

  /* main window */
  rack->main_window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title (GTK_WINDOW (rack->main_window), PACKAGE_NAME);
  gtk_window_set_default_size (GTK_WINDOW (rack->main_window), 620, 460);
  g_signal_connect (G_OBJECT (rack->main_window), "delete_event",
                        G_CALLBACK (window_destroy_callback), rack);  
                        
  

  /* button/viewport box */
  main_box = gtk_vbox_new (FALSE, 0);
  gtk_widget_show (main_box);
  gtk_container_add (GTK_CONTAINER (rack->main_window), main_box);

  jack_rack_init_gui_menu (rack, main_box);


  /* top button box and buttons */
  toolbar_handle = gtk_handle_box_new ();
  gtk_widget_show (toolbar_handle);
  gtk_box_pack_start (GTK_BOX (main_box), toolbar_handle, FALSE, FALSE, 0);

  toolbar = gtk_toolbar_new ();
  gtk_widget_show (toolbar);
  gtk_container_add (GTK_CONTAINER (toolbar_handle), toolbar);
  gtk_toolbar_set_icon_size (GTK_TOOLBAR (toolbar), GTK_ICON_SIZE_SMALL_TOOLBAR);


  rack->add = gtk_toolbar_insert_stock (GTK_TOOLBAR (toolbar),
                                        GTK_STOCK_ADD,
                                        "Add a plugin",
                                        NULL, NULL, NULL, -1);
  menu = plugin_mgr_get_menu (rack->plugin_mgr, G_CALLBACK (add_callback), rack);
  g_signal_connect_swapped (G_OBJECT (rack->add), "event",
                            G_CALLBACK (plugin_button_callback),
                            G_OBJECT (menu));
                                                         
  gtk_toolbar_append_space (GTK_TOOLBAR (toolbar));

  gtk_toolbar_insert_stock (GTK_TOOLBAR (toolbar),
                            GTK_STOCK_NEW,
                            "Clear the rack",
                            NULL, G_CALLBACK (new_callback), rack, -1);

  gtk_toolbar_insert_stock (GTK_TOOLBAR (toolbar),
                            GTK_STOCK_OPEN,
                            "Open a rack configuration file",
                            NULL, G_CALLBACK (open_callback), rack, -1);

  gtk_toolbar_insert_stock (GTK_TOOLBAR (toolbar),
                            GTK_STOCK_SAVE,
                            "Save the rack configuration to the current file",
                            NULL, G_CALLBACK (save_callback), rack, -1);

  gtk_toolbar_insert_stock (GTK_TOOLBAR (toolbar),
                            GTK_STOCK_SAVE_AS,
                            "Save the rack configuration to a file",
                            NULL, G_CALLBACK (save_as_callback), rack, -1);

#ifdef HAVE_LADCCA
  if (cca_enabled (global_cca_client))
    {
      gtk_toolbar_append_space (GTK_TOOLBAR (toolbar));

      rack->cca_save = gtk_toolbar_insert_stock (GTK_TOOLBAR (toolbar),
                                           GTK_STOCK_SAVE,
                                           "Save the LADCCA project",
                                           NULL, G_CALLBACK (cca_save_callback), rack, -1);
    }

#endif
  gtk_toolbar_append_space (GTK_TOOLBAR (toolbar));

  gtk_toolbar_insert_stock (GTK_TOOLBAR (toolbar),
                            GTK_STOCK_QUIT,
                            "Quit JACK Rack",
                            NULL, G_CALLBACK (quit_callback), rack, -1);


                              

  /* viewport thingy */
  plugin_scroll = gtk_scrolled_window_new (NULL, NULL);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (plugin_scroll),
                                  GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_box_pack_start (GTK_BOX (main_box), plugin_scroll, TRUE, TRUE, 0);
  gtk_widget_show (plugin_scroll);

  plugin_viewport = gtk_viewport_new (NULL, NULL);
  gtk_container_add (GTK_CONTAINER (plugin_scroll), plugin_viewport);
  gtk_widget_show (plugin_viewport);


  /* the main vbox for the plugins */
  rack->plugin_box = gtk_vbox_new (FALSE, 0);
  gtk_container_add (GTK_CONTAINER (plugin_viewport), rack->plugin_box);
  gtk_widget_show (rack->plugin_box);

  gtk_widget_show (rack->main_window);

  gtk_idle_add (idle_callback, rack);
}

void
jack_rack_init_splash_screen (jack_rack_t * jack_rack)
{
  gchar * logo_file;
  GtkWidget * logo;
  GtkWidget * box;
  
  jack_rack->splash_screen = gtk_window_new (GTK_WINDOW_TOPLEVEL);
  gtk_widget_realize (jack_rack->splash_screen);
  gtk_window_set_decorated (GTK_WINDOW (jack_rack->splash_screen), FALSE);
  gtk_window_set_resizable (GTK_WINDOW (jack_rack->splash_screen), FALSE);
  gtk_container_set_resize_mode (GTK_CONTAINER (jack_rack->splash_screen), GTK_RESIZE_IMMEDIATE);
  gtk_window_set_position (GTK_WINDOW (jack_rack->splash_screen), GTK_WIN_POS_CENTER_ALWAYS);
  
  box = gtk_vbox_new (FALSE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (box), 3);
  gtk_widget_show (box);
  gtk_container_add (GTK_CONTAINER (jack_rack->splash_screen), box);
  
  logo_file = g_strdup_printf ("%s/%s", PKGDATADIR, JACK_RACK_LOGO_FILE);
  logo = gtk_image_new_from_file (logo_file);
  gtk_widget_show (logo);
  gtk_box_pack_start (GTK_BOX (box), logo, FALSE, FALSE, 0);
  
  jack_rack->splash_screen_text = gtk_label_new (NULL);
  gtk_widget_show (jack_rack->splash_screen_text);
  gtk_box_pack_start (GTK_BOX (box), jack_rack->splash_screen_text, TRUE, TRUE, 0);
  gtk_misc_set_alignment (GTK_MISC (jack_rack->splash_screen_text), 0.0, 0.5);
  
  gtk_widget_show (jack_rack->splash_screen);

  for (int i = 0; i < 20; i++)
    gtk_main_iteration_do (FALSE);
}

jack_rack_t *
jack_rack_new (const char * jack_client_name)
{
  jack_rack_t *rack;

  rack = g_malloc (sizeof (jack_rack_t));
  rack->splash_screen = NULL;
  rack->splash_screen_text = NULL;
  rack->slots = NULL;
  rack->slots_end = NULL;
  rack->slot_count = 0;
  rack->filename = NULL;
  rack->saved_settings = NULL;
  rack->shutdown = FALSE;
  rack->state = STATE_NORMAL;
#ifdef HAVE_LADCCA  
  rack->cca_save = NULL;
#endif

  /* init the fifos */
  rack->gui_to_process = lff_new (32, sizeof (ctrlmsg_t));
  rack->process_to_gui = lff_new (32, sizeof (ctrlmsg_t));
  
  jack_rack_init_splash_screen (rack);
  
  rack->process_info = process_info_new (rack, jack_client_name);
  rack->plugin_mgr = plugin_mgr_new (rack);

  rack->gui_to_process = rack->process_info->gui_to_process;
  rack->process_to_gui = rack->process_info->process_to_gui;

  jack_rack_init_gui (rack);
  
  gtk_widget_destroy (rack->splash_screen);
  rack->splash_screen = NULL;

  return rack;
}


void
jack_rack_destroy (jack_rack_t * jack_rack)
{
  gtk_widget_destroy (jack_rack->main_window);
  
  process_info_destroy (jack_rack->process_info);
  plugin_mgr_destroy (jack_rack->plugin_mgr);
  
  lff_destroy (jack_rack->gui_to_process);
  lff_destroy (jack_rack->process_to_gui);
  
  g_free (jack_rack);
}

plugin_t *
jack_rack_instantiate_plugin (jack_rack_t * jack_rack, plugin_desc_t * desc)
{
  plugin_t * plugin;
  
  /* check whether or not the plugin is RT capable and confirm with the user if it isn't */
  if (!LADSPA_IS_HARD_RT_CAPABLE(desc->properties)) {
    GtkWidget * dialog;
    gint response;

    dialog = gtk_message_dialog_new (GTK_WINDOW(jack_rack->main_window),
               GTK_DIALOG_DESTROY_WITH_PARENT|GTK_DIALOG_MODAL,
               GTK_MESSAGE_WARNING,
               GTK_BUTTONS_YES_NO,
               "Plugin not RT capable\n\nThe plugin '%s' does not describe itself as being capable of real-time operation.  You may experience drop outs or jack may even kick us out if you use it.\n\nAre you sure you want to add this plugin?",
               desc->name);
    response = gtk_dialog_run (GTK_DIALOG (dialog));
    gtk_widget_destroy (dialog);
    
    if (response != GTK_RESPONSE_YES) {
      return NULL;
    }
  }

  /* create the plugin */
  plugin = plugin_new (desc);

  if (!plugin) {
    GtkWidget * dialog;
    
    dialog = gtk_message_dialog_new (GTK_WINDOW(jack_rack->main_window),
               GTK_DIALOG_DESTROY_WITH_PARENT,
               GTK_MESSAGE_ERROR,
               GTK_BUTTONS_CLOSE,
               "Error loading file plugin '%s' from file '%s'",
               desc->name, desc->object_file);
    gtk_dialog_run (GTK_DIALOG (dialog));
    gtk_widget_destroy (dialog);
    return NULL;
  }
  
  return plugin;
}

void
jack_rack_add_plugin (jack_rack_t * jack_rack, plugin_desc_t * desc)
{
  plugin_t * plugin;
  ctrlmsg_t ctrlmsg;
  
  plugin = jack_rack_instantiate_plugin (jack_rack, desc);
  
  if (!plugin)
    return;
  
  /* send the chain link off to the process() callback */
  ctrlmsg.type = CTRLMSG_ADD;
  ctrlmsg.pointer = plugin;
  ctrlmsg.second_pointer = desc;
  lff_write (jack_rack->gui_to_process, &ctrlmsg);
}


void
jack_rack_add_plugin_slot (jack_rack_t * jack_rack, plugin_t * plugin)
{
  plugin_slot_t *plugin_slot;
  settings_t * settings = NULL;
  GSList * list;
  
  /* see if there's any saved settings that match the plugin id */
  for (list = jack_rack->saved_settings; list; list = g_slist_next (list))
    {
      settings = (settings_t *) list->data;
      if (settings->desc->id == plugin->desc->id)
        {
          jack_rack->saved_settings = g_slist_remove (jack_rack->saved_settings, settings);
          break;
        }
      settings = NULL;
    }

  /* create the plugin_slot */
  plugin_slot = plugin_slot_new (jack_rack, plugin, settings);
  
  if (settings)
    settings_destroy (settings);

  if (jack_rack->slots)
    {
      plugin_slot->prev = jack_rack->slots_end;
      plugin_slot->prev->next = plugin_slot;
      plugin_slot->next = NULL;
      jack_rack->slots_end = plugin_slot;
      plugin_slot->index = plugin_slot->prev->index + 1;
    }
  else
    {
      plugin_slot->next = NULL;
      plugin_slot->prev = NULL;
      jack_rack->slots = plugin_slot;
      jack_rack->slots_end = plugin_slot;
      plugin_slot->index = 0;
    }
  jack_rack->slot_count++;

  gtk_box_pack_start (GTK_BOX (jack_rack->plugin_box),
                      plugin_slot->main_vbox, FALSE, FALSE, 0);
}

void
jack_rack_remove_plugin_slot (jack_rack_t * jack_rack, plugin_slot_t * plugin_slot)
{
  /* adjust pointers */
  if (plugin_slot->prev)
    {
      plugin_slot->prev->next = plugin_slot->next;
    }
  else
    {
      jack_rack->slots = plugin_slot->next;
    }

  /* adjust pointers and indicies */
  if (plugin_slot->next)
    {
      plugin_slot_t *ps;

      plugin_slot->next->prev = plugin_slot->prev;

      for (ps = plugin_slot->next; ps; ps = ps->next)
        {
          ps->index--;
        }
    }
  else
    {
      jack_rack->slots_end = plugin_slot->prev;
    }
  jack_rack->slot_count--;

  plugin_slot_destroy (plugin_slot);
}

void
jack_rack_move_plugin_slot (jack_rack_t * jack_rack, plugin_slot_t * plugin_slot, gint up)
{
  plugin_slot_t *tmpslot;
  if (up)
    {
      if (!plugin_slot->prev)
        return;

      plugin_slot->index--;
      plugin_slot->prev->index++;

      if (plugin_slot->next)
        {
          plugin_slot->next->prev = plugin_slot->prev;
        }
      else
        {
          jack_rack->slots_end = plugin_slot->prev;
        }
      tmpslot = plugin_slot->prev->prev;
      if (tmpslot)
        {
          tmpslot->next = plugin_slot;
        }
      else
        {
          jack_rack->slots = plugin_slot;
        }
      plugin_slot->prev->next = plugin_slot->next;
      plugin_slot->prev->prev = plugin_slot;
      plugin_slot->next = plugin_slot->prev;
      plugin_slot->prev = tmpslot;

    }
  else
    {
      if (!plugin_slot->next)
        return;

      plugin_slot->index++;
      plugin_slot->next->index--;

      if (plugin_slot->prev)
        {
          plugin_slot->prev->next = plugin_slot->next;
        }
      else
        {
          jack_rack->slots = plugin_slot->next;
        }
      tmpslot = plugin_slot->next->next;
      if (tmpslot)
        {
          tmpslot->prev = plugin_slot;
        }
      else
        {
          jack_rack->slots_end = plugin_slot;
        }
      plugin_slot->next->prev = plugin_slot->prev;
      plugin_slot->next->next = plugin_slot;
      plugin_slot->prev = plugin_slot->next;
      plugin_slot->next = tmpslot;
    }

  gtk_box_reorder_child (GTK_BOX (jack_rack->plugin_box),
                         plugin_slot->main_vbox, plugin_slot->index);
}

void
jack_rack_change_plugin_slot (jack_rack_t * jack_rack, plugin_slot_t * plugin_slot, plugin_t * plugin)
{
  plugin_slot_change_plugin (plugin_slot, plugin);
}

void
jack_rack_clear_plugins (jack_rack_t * jack_rack, plugin_t * plugin)
{
  plugin_t * next_plugin;
  plugin_slot_t * slot;
  
  for (; plugin; plugin = next_plugin)
    {
      next_plugin = plugin->next;
      
      for (slot = jack_rack->slots; slot; slot = slot->next)
        {
          if (slot->plugin == plugin)
            {
              jack_rack_remove_plugin_slot (jack_rack, slot);
              plugin_destroy (plugin);
              break;
            }
        }
      
    }
}

plugin_slot_t *
jack_rack_get_plugin_slot (jack_rack_t * jack_rack, unsigned long plugin_index)
{
  plugin_slot_t * slot;
  
  for (slot = jack_rack->slots; plugin_index && slot; slot = slot->next, plugin_index--);
  
  return slot;
}

void
jack_rack_display_error (jack_rack_t * jack_rack, const char * message)
{
  static GtkWidget * dialog = NULL;
  
  if (!dialog)
    dialog = gtk_message_dialog_new (GTK_WINDOW (jack_rack->main_window),
                                     GTK_DIALOG_DESTROY_WITH_PARENT|GTK_DIALOG_MODAL,
                                     GTK_MESSAGE_ERROR,
                                     GTK_BUTTONS_OK,
                                     "%s", message);

  gtk_dialog_run (GTK_DIALOG(dialog));
  gtk_widget_hide (dialog);
}

void
jack_rack_set_filename (jack_rack_t * jack_rack, const char * filename)
{
  const char * base_filename;
  char * window_title;
  
  if (jack_rack->filename == filename)
    return;
  
  set_string_property (jack_rack->filename, filename);

  base_filename = strrchr (filename, '/');
  if (base_filename)
    base_filename++;
  else
    base_filename = filename;
    
  window_title = g_strdup_printf ("%s - %s", PACKAGE_NAME, base_filename);
  gtk_window_set_title (GTK_WINDOW (jack_rack->main_window), window_title);
  g_free (window_title);
}


void
jack_rack_set_state    (jack_rack_t * jack_rack, jack_rack_state_t state)
{
  jack_rack->state = state;
}

jack_rack_state_t
jack_rack_get_state (jack_rack_t * jack_rack)
{
  return jack_rack->state;
}



/* EOF */



